/*Please do not edit this file, it contains extracted strings from GUI version.*/
#include <stdio.h>
void main(void)
{
puts(_("WP2LaTeX"));
puts(_("Files"));
puts(_("Settings"));
puts(_("Packages"));
puts(_("Info"));
puts(_("Wasy"));
puts(_("AMS Symb"));
puts(_("BBM"));
puts(_("RSFS"));
puts(_("Fancy HDR"));
puts(_("Latex Sym"));
puts(_("Make idx"));
puts(_("Long Table"));
puts(_("CP styles"));
puts(_("Fontenc[T1]"));
puts(_("TIPA"));
puts(_("Hebrew support"));
puts(_("pifont"));
puts(_("graphics"));
puts(_("Graphic X"));
puts(_("Input PS"));
puts(_("EpsFig"));
puts(_("ulem"));
puts(_("cyrillic"));
puts(_("color"));
puts(_("accents"));
puts(_("Scalerel"));
puts(_("mxedruli"));
puts(_("ArabTeX"));
puts(_("Input File"));
puts(_("Output File"));
puts(_("Log File"));
puts(_("Change"));
puts(_("Change"));
puts(_("Change"));
puts(_("Convert"));
puts(_("Exit"));
puts(_("Output Codepage"));
puts(_("WP character set"));
puts(_("LaTeX version"));
puts(_("Image Path (please prefer relative paths):"));
puts(_("Force document format"));
puts(_("Default character set"));
puts(_("Ignore TABs"));
puts(_("Extract Images"));
puts(_("Safe Mode"));
puts(_("Silent Mode"));
puts(_("TeX chars"));
puts(_("Fix Spaces"));
puts(_("Shrink oversized images"));
puts(_("Available Convertor Modules:"));
puts(_("Supported Image formats:"));
puts(_("https://ftsoft.com.cz/wp2latex/wp2latex.htm"));
puts(_("&Help"));
puts(_("(C)  1996-2025 Jaroslav Fojtik"));
puts(_("&Splash"));
puts(_("WP2LaTeX messages"));
puts(_("Conversion Results"));
puts(_("Conversion from"));
puts(_("Label2"));
puts(_("To"));
puts(_("Label4"));
puts(_("Finished successfully!"));
puts(_("OK"));
puts(_("Dialog"));
puts(_("Licence for use is royalty free granted."));
puts(_("Prague   ---- Made by J.Fojtik     ---- Date : 2001-2025"));
puts(_("Shareware version:"));
puts(_("----"));
puts(_(">>WP2LaTeX<<< Conversion program: From Wordperfect to LaTeX Version"));
puts(_("Alternative URL: http://78.108.103.11/~fojtik/wp2latex/wp2latex.htm"));
puts(_("or: http://ftsoft.wz.cz/wp2latex/wp2latex.htm"));
puts(_("Get latest version from: http://ftsoft.com.cz/wp2latex/wp2latex.htm"));
puts(_("OK"));
puts(_("WP2LaTeX"));
puts(_("Enter Password:"));
puts(_("OK"));
puts(_("Progress"));
puts(_("Label1"));
puts(_("WP2LaTeX - Conversion progress"));
puts(_("WP2LaTeX help"));
puts(_("OK"));
puts(_("WP2LaTeX copyright"));
puts(_("OK"));
puts(_("Frm_WaitForJobs"));
puts(_("Waiting for asynchronous threads to finish."));
puts(_("&Ignore"));
}
