#include <io.h>
#include <fcntl.h>
#include <Windows.h>
#include <WinCon.h>

#include <stdio.h>


void FixConsole2UTF8(void)
{
  SetConsoleOutputCP(CP_UTF8);
  //_setmode(_fileno(stdout), _O_U8TEXT);
  //printf("Konverze přepnuta na UTF8!");
}