
#define APPVERSION    4
#define APPREVISION   14
#define APPRELEASE    20250228

#define VERSIONDESCRIPTION          "Conversion from WP (and more other formats) to LaTeX"
#define VERSIONCOMPANYNAME          "Jaroslav Fojtik"
#define VERSIONPRODUCTNAME          "WP2LaTeX"
//#define VERSIONSTR                  "4.5.0"
#define VERSIONCOPYRIGHT            "Copyright (C) Jaroslav Fojtik 1996-2025"
#define VERSIONNAME                 "WP2LaTeX.exe"

#define VERSIONTYPE                 VFT_APP
#define VERSIONSUBTYPE              VFT2_UNKNOWN
#define VERSIONFILEFLAGSMASK        VS_FF_DEBUG
#ifdef DEBUG
#define VERSIONFLAGS                VS_FF_DEBUG
#else
#define VERSIONFLAGS                0
#endif


#ifndef DEFINE_ONLY

#include <windows.h>
#include <ntverp.h>

#include "verinfo.ver"

/////////////////////////////////////////////////////

wp2latex    ICON   "../bin/win/wp2latex.ico"


////////////////////////////////////////////

//#define wxUSE_NO_MANIFEST 1
//#include "wx/msw/wx.rc" 

#endif  // DEFINE_ONLY